# Add the current path to python
import sys
sys.path.append('.')
import traceback
# Required if you need to output gif
import matplotlib
matplotlib.use('Agg')
from matplotlib import rcParams
rcParams.update({'figure.autolayout': True})
font_size = 18
matplotlib.rc('xtick', labelsize=font_size) 
matplotlib.rc('ytick', labelsize=font_size) 

import argparse
import numpy as np
import torch
import util_torch
import datetime
from model_fast_thinking import IOC_Cooperative

def main_basic(opt):

    if opt.category == "default": 
        opt.category = datetime.datetime.now().strftime('%b%d_%H-%M-%S') 
    if opt.mode != "test": 
        opt.category = "_".join(["FS", opt.des_net, opt.gen_net, opt.category])
    torch.manual_seed(opt.seed)
    np.random.seed(opt.seed)

    torch.set_default_tensor_type(torch.cuda.FloatTensor if opt.device=='cuda' else torch.FloatTensor)
    opt.future_size = opt.traj_size - opt.his_size

    default_sizes = {
        'Full': (1000, 100000, 10000), 'Half': (100, 100000, 10000),
        'Big': (10, 5000, 1000), 'Small': (100, 100, 100),
        'Tiny': (5, 5, 5), 'One': (1, 1, 1)}
    if opt.sizes in default_sizes:
        opt.batch_size, opt.train_size, opt.test_size = default_sizes[opt.sizes]
    if opt.mode == "test":
        return main_test(opt)

    # if opt.continue_train:
    #     with open("output/%s/config.pkl" % opt.category, "rb") as conf:
    #         opt = pickle.load(conf)
    #     print("Load config.")
    #     print(opt)
    #     opt.continue_train = 1
    
    train_data = getattr(util_torch, "%sDataset" %("Single" if opt.single else "Multi"))(opt, "train")
    test_data = getattr(util_torch, "%sDataset" %("Single" if opt.single else "Multi"))(opt, "test")
    print("Network selection: %s" % opt.des_net)
    try: 
        model = IOC_Cooperative(opt)
        model.train(train_data, test_data)
    except Exception as e:
        print(traceback.format_exc()) 
        # util_torch.PushNotification().post_text("[CIOC] FAILED in %s! Error message: %s" % (opt.category, str(traceback.format_exc())))

def arg_parse(): 

    parser = argparse.ArgumentParser()
    # training hyper-parameters
    parser.add_argument('-num_epochs', type=int, default=1000)
    parser.add_argument('-batch_size', type=int, default=1000)
    parser.add_argument('-agent_size', type=int, default=64)
    parser.add_argument('-train_size', type=int, default=100)
    parser.add_argument('-hidden_size', type=int, default=64)
    parser.add_argument('-test_size', type=int, default=10000)
    parser.add_argument('-latent_size', type=int, default=4, help='number of frames used in training data')
    parser.add_argument('-num_layer', type=int, default=2, help='number of frames used in training data')
    parser.add_argument('-num_gen_iter', type=int, default=1, help='number of frames used in training data')


    parser.add_argument('-traj_size', type=int, default=31, help='number of frames used in training data')
    parser.add_argument('-his_size', type=int, default=1, help='number of frames used in training data')
    parser.add_argument('-feature_size', type=int, default=10)

    parser.add_argument('-lr_des', type=float, default=1e-1, help='learning rate for discriptor')
    parser.add_argument('-lr_gen', type=float, default=2e-3, help='learning rate for generator')
    parser.add_argument('-lr_gen_c', type=float, default=5e-4, help='learning rate for generator critic if required')
    parser.add_argument('-weight_decay', type=float, default=0, help='learning rate for generator')
    parser.add_argument('-gamma_decay', type=float, default=0.999, help='learning rate for generator')
    parser.add_argument('-beta1', type=float, default=0.5, help='momentum1 in Adam')
    parser.add_argument('-delta', type=float, default=0.1)
    parser.add_argument('-sample_steps', type=int, default=64)
    parser.add_argument('-num_sample', type=int, default=1)
    parser.add_argument('-noise_steps', type=int, default=1024)
    parser.add_argument('-default_limit', type=float, default=0.1)
    parser.add_argument('-mode', type=str, default="")

    parser.add_argument('-neg', type=int, default=0, help='Restrict negative theta')
    parser.add_argument('-des_net', type=str, default='Linear')
    parser.add_argument('-gen_net', type=str, default='None')
    parser.add_argument('-cold_start', type=int, default=1, help='1 = use zero control, 0 = use constant control')
    parser.add_argument('-costFunc', type=str, default='DefaultISEE')
    parser.add_argument('-category', type=str, default='default')
    parser.add_argument('-single', type=int, default=1)
    parser.add_argument('-device', type=str, default="cuda")
    parser.add_argument('-sizes', type=str, default='Full')
    parser.add_argument('-data', type=str, default='ISEE', help='root directory of data')
    parser.add_argument('-normalize', type=int, default=1)
    parser.add_argument('-seed', type=int, default=666, help='number of steps to output synthesized image')
    parser.add_argument('-dynamic', type=str, default='Reparam', help='Dynamic Model')
    parser.add_argument('-sigma', type=float, default=0.1)  # [0.07, 0.005])
    parser.add_argument('-std_para', type=float, default=1)  # [0.07, 0.005])
    parser.add_argument('-print_level', type=int, default=50)
    parser.add_argument('-eval_interval', type=int, default=20)
    parser.add_argument('-continue_train', type=int, default=0)
    
    return parser.parse_args()

if __name__ == '__main__':
    
    main_basic(arg_parse())
    

