import numpy as np
import pickle
import time
import csv
import os
import torch.utils.data as tdata
import torch
from scipy.stats import norm
import matplotlib.pyplot as plt
from matplotlib import animation, rc
from torch.utils.tensorboard import SummaryWriter

class NGSIMSingleDataset(tdata.Dataset):
    """Face Landmarks dataset."""

    def __init__(self, traj_size=10000, path="output_multi.pkl", max_traj=100000000, device='cpu', 
                 use_repara=False, shrink=True):

        data = np.load(path)

        # Modify other vehicle:

        # x_l, y_l, null = 150, 50, 0
        # for entry in data:
        #     for d in entry:
        #         # 1. Front vehicle missing
        #         if d[8] == null:
        #             d[8], d[9] = d[0] + x_l, d[1]
        #         # 2. Back vehicle missing
        #         if d[12] == null:
        #             d[12], d[13] = d[0] - x_l, d[1]
        #         # 3. Front-left vehicle missing
        #         if d[16] == null:
        #             d[16], d[17] = d[0] + x_l, d[1] - y_l
        #         # 4. Front-right vehicle missing
        #         if d[20] == null:
        #             d[20], d[21] = d[0] + x_l, d[1] + y_l
        #         # 5. Back-left vehicle missing
        #         if d[24] == null:
        #             d[24], d[25] = d[0] - x_l, d[1] - y_l
        #         # 6. Back-right vehicle missing
        #         if d[28] == null:
        #             d[28], d[29] = d[0] - x_l, d[1] + y_l
        # print('Modification made in other vehicle, please use \'easy\' costFunc. ')

        status = data[:max_traj, 1:traj_size+1, :8]
        if use_repara:
            status[..., 6] = status[..., 6] - status[..., 4]
            status[..., 7] = status[..., 7] - status[..., 5]
            print("Use delta control")
        if shrink:
            self.multiple_factor = [1 / np.sqrt(np.var(status[...,6])), 1 / np.sqrt(np.var(status[...,7]))]
            print(self.multiple_factor)
            status[..., 4] *= self.multiple_factor[0]
            status[..., 5] *= self.multiple_factor[1]
            status[..., 6] *= self.multiple_factor[0]
            status[..., 7] *= self.multiple_factor[1]
        else:
            self.multiple_factor = [1, 1]

        assert traj_size < data.shape[1]

        self.status = torch.tensor(status, dtype=torch.float32, device=device)
        self.extend = torch.tensor(data[:max_traj, 1:traj_size+1, 8:], dtype=torch.float32, device=device)
        self.state_size, self.action_size, self.extend_size = 6, 2, self.extend.shape[-1]
        print('Data loaded', self.status.shape, self.extend.shape)

    def __len__(self):
        return len(self.status)

    def __getitem__(self, idx):
        return {'status': self.status[idx, ...], 'extend': self.extend[idx, ...]}

class NGSIMSingleDatasetD(tdata.Dataset):
    """Face Landmarks dataset."""

    def __init__(self, data, tensor=torch.cuda.FloatTensor):
        self.status = data[:, :, :8]
        self.extend = data[:, :, 8:]
        self.tensor = tensor

    def __len__(self):
        return len(self.status)

    def __getitem__(self, idx):
        return {'status': self.tensor(self.status[idx, ...]), 'extend': self.tensor(self.extend[idx, ...])}

class NGSIMMultiDataset(tdata.Dataset):
    """Face Landmarks dataset."""

    def __init__(self, traj_size=10000, agent_size_max=64, use_original=True,
                 path="output_multi.pkl", split=4, max_traj=100000000, device='cpu', 
                 use_repara=False, shrink=True):
        with open(path, 'rb') as pickle_file:

            data = pickle.load(pickle_file)
            data = data[:max_traj]
            n = len(data)
            trajs_size = data[0]['x_new'].shape[1]
            fea_size = data[0]['x_new'].shape[2]
            ext_size = data[0]['x_extend'].shape[1]
            max_length = np.max([d['x_new'].shape[0] for d in data])

            status = np.zeros((n * split, max(agent_size_max, max_length), trajs_size, fea_size))
            extend = np.zeros((n * split, max(agent_size_max, max_length), ext_size + 1))
            overall_th = 0
            for idx, d in enumerate(data):
                agent_size = d['x_new'].shape[0]
                curr_status = d['x_ori'] if use_original else d['x_new']
                range_pnt = np.linspace(np.min(curr_status[:, 0, 0], axis=0),
                                        np.max(curr_status[:, 0, 0], axis=0), split+1)
                sorted_pnt = curr_status[:, 0, 0].argsort()

                if agent_size < 50:
                    for i in range(agent_size):
                        while np.sum(np.isnan(curr_status[i])) > 0:
                            curr_status[i] = curr_status[agent_size-1]
                            d['x_extend'][i] = d['x_extend'][agent_size-1]
                            agent_size -= 1
                    status[overall_th, :agent_size, ...] = curr_status[:agent_size]
                    extend[overall_th, :agent_size, :ext_size] = d['x_extend'][:agent_size]
                    extend[overall_th, :agent_size, ext_size] = 1
                    overall_th += 1
                else:
                    ith_range = 0
                    cnt_number = 0
                    for ith in sorted_pnt:
                        if curr_status[ith, 0, 0] <= range_pnt[ith_range + 1]:
                            if np.sum(np.isnan(curr_status[ith, ...])) == 0:
                                status[overall_th, cnt_number, ...] = curr_status[ith, ...]
                                extend[overall_th, cnt_number, :ext_size] = d['x_extend'][ith, ...]
                                cnt_number += 1
                        else:
                            extend[overall_th, :cnt_number, ext_size] = 1
                            ith_range += 1
                            overall_th += 1
                            cnt_number = 0
                    extend[overall_th, :cnt_number, ext_size] = 1
                    overall_th += 1

        # Updated 2019.05.13:
        # status[..., 1:, 4:6] = (status[..., 1:, 2:4] - status[..., :-1, 2:4])*10
        # status[..., 4:8] = 0.001

        if traj_size < status.shape[2]:
            status = status[:, :, 1:(traj_size+1), :]
        overall_th = min(overall_th, max_traj)
        
        if use_repara:
            status[..., 6] = status[..., 6] - status[..., 4]
            status[..., 7] = status[..., 7] - status[..., 5]
            print("Use delta control")
        if shrink:
            self.multiple_factor = [1 / np.sqrt(np.var(status[...,6])), 1 / np.sqrt(np.var(status[...,7]))]
            print(self.multiple_factor)
            status[..., 4] *= self.multiple_factor[0]
            status[..., 5] *= self.multiple_factor[1]
            status[..., 6] *= self.multiple_factor[0]
            status[..., 7] *= self.multiple_factor[1]
        else:
            self.multiple_factor = [1, 1]
        self.status = torch.tensor(status[:overall_th, :agent_size_max], dtype=torch.float32, device=device)
        self.extend = torch.tensor(extend[:overall_th, :agent_size_max], dtype=torch.float32, device=device)

        print('Data loaded', self.status.shape, self.extend.shape)

    def __len__(self):
        return len(self.status)

    def __getitem__(self, idx):
        return {'status': self.status[idx, ...], 'extend': self.extend[idx, ...]}

class CARLAMultiDataset(tdata.Dataset):
    """Face Landmarks dataset."""

    def __init__(self, traj_size=10000, agent_size_max=64, path="output_v1_control.npy", max_traj=11, device='cuda'):

        data = np.load(path)
        data = data[:max_traj, :agent_size_max, :traj_size]
        shape = data.shape
        status = np.zeros((shape[0], shape[1], shape[2], 8))
        # x, y, v, theta, 0, 0, u1, u2
        extend = np.ones((shape[0], shape[1], shape[2], 4))
        # goal, limit, exist
        status[..., :4] = data[..., [1,0,3,2]]
        goal = data[..., -1, 0:2]
        extend[..., 0:2] = np.repeat(goal[:, :, np.newaxis, :], shape[2], axis=2)
        extend[..., 2] = data[..., 8] / 3.6
        self.status = torch.tensor(status, dtype=torch.float32, device=device)
        self.extend = torch.tensor(extend, dtype=torch.float32, device=device)
        print('Data loaded', self.status.shape, self.extend.shape)

    def __len__(self):
        return len(self.status)

    def __getitem__(self, idx):
        return {'status': self.status[idx, ...], 'extend': self.extend[idx, ...]}

class SingleDataset(tdata.Dataset):

    def __init__(self, config, split="train"):

        data = np.load("data/%s_%s.npy" % (config.data, split))
        config, data = getattr(self, "p_%s" % config.data)(config, data)
        s, a, e = config.state_size, config.action_size, config.extend_size
        data = data[:config.train_size, 2:2+config.traj_size]

        if split == "train":
            np.random.shuffle(data) 

        if config.dynamic == 'Reparam':
            data[..., s:s+a] = data[..., s:s+a] - data[..., s-a:s]

        if config.normalize: 

            if not hasattr(config, "multiple_factor"):
                config.multiple_factor = 1 / np.sqrt(data[:, :, s:s+a].var(axis=(0,1)))
                print("Normalize controls to: ", config.multiple_factor)
                # config.multiple_factor = [1 / np.percentile(train_data[:, :, 6], 95), 1 / np.percentile(train_data[:, :, 7], 95)]
                # print("---", config.multiple_factor, np.min(train_data[:, :, 6]), np.max(train_data[:, :, 6]), np.min(train_data[:, :, 7]), np.max(train_data[:, :, 7]))
            
            data[..., s:s+a] *= config.multiple_factor
            # The last part of the state is previous control.
            data[..., s-a:s] *= config.multiple_factor
        else: 
            config.multiple_factor = np.ones((2))
            
        self.status = data[:, :, :s+a]
        self.extend = data[:, :, s+a:]
        self.tensor = torch.FloatTensor if  config.device=="cpu" else torch.cuda.FloatTensor
        print('Data loaded', self.status.shape, self.extend.shape)

    def p_ISEE(self, config, data):

        config.state_size, config.action_size, config.extend_size = 6, 2, 29
        data = data[:, 1:]
        return config, data

    def p_NGSIM(self, config, data):
                
        config.state_size, config.action_size, config.extend_size = 6, 2, 47
        data = data[:, 1:]
        return config, data

    def p_ETH(self, config, data):
        
        config.state_size, config.action_size, config.extend_size = 2, 2, 0
        return config, data

    def __len__(self):
        return len(self.status)

    def __getitem__(self, idx):
        return {'status': self.tensor(self.status[idx, ...]), 'extend': self.tensor(self.extend[idx, ...])}

def load_data(opt):

    if opt.data == "ISEE":
        
        date = [1, 4, 5, 7, 8, 9, 15, 21, 22, 12, 19]
        opt.data_path = ['data/package_%02d.npy' % d for d in date]
        np.random.seed(opt.seed)
        train_data = np.load(opt.data_path[0])
        for path in opt.data_path[1:-2]:
            train_data = np.concatenate([train_data, np.load(path)])
            if train_data.shape[0] > opt.train_size:
                break
            
        train_data = train_data[:opt.train_size, 1:opt.traj_size+1, :]
        np.random.shuffle(train_data)
        test_data = np.load(opt.data_path[-2])
        if test_data.shape[0] < opt.test_size:
            test_data = np.concatenate([test_data, np.load(opt.data_path[-1])])
        test_data = test_data[:opt.test_size, 1:opt.traj_size+1, :]
        if opt.dynamic == 'Reparam':
            train_data[..., 6:8] = train_data[..., 6:8] - train_data[..., 4:6]
            test_data[..., 6:8] = test_data[..., 6:8] - test_data[..., 4:6]
        print(train_data.shape)
        opt.multiple_factor = multiple_factor = [1 / np.sqrt(np.var(train_data[:, :, 6])), 1 / np.sqrt(np.var(train_data[:, :, 7]))]
        # opt.multiple_factor = multiple_factor = [1 / np.percentile(train_data[:, :, 6], 95), 1 / np.percentile(train_data[:, :, 7], 95)]
        print("---", multiple_factor, np.min(train_data[:, :, 6]), np.max(train_data[:, :, 6]), np.min(train_data[:, :, 7]), np.max(train_data[:, :, 7]))
        train_data[..., 4] *= multiple_factor[0]
        train_data[..., 6] *= multiple_factor[0] 
        test_data[..., 4] *= multiple_factor[0]
        test_data[..., 6] *= multiple_factor[0]
        train_data[..., 5] *= multiple_factor[1]
        train_data[..., 7] *= multiple_factor[1]
        test_data[..., 5] *= multiple_factor[1]
        test_data[..., 7] *= multiple_factor[1]
        train_data = NGSIMSingleDatasetD(train_data, torch.FloatTensor if opt.device=="cpu" else torch.cuda.FloatTensor)
        test_data = NGSIMSingleDatasetD(test_data, torch.FloatTensor if opt.device=="cpu" else torch.cuda.FloatTensor)
        # opt.multiple_factor = [1,1]
        opt.state_size, opt.action_size, opt.extend_size = 6, 2, 29
    
    elif opt.data == "NGSIM":

        # dataset = NGSIMSingleDataset(opt.traj_size, '../../Dataset/ngsim/package_multi_compact.npy', opt.train_size, device=opt.device, 
        # use_repara=(opt.dynamic=='Reparam'), shrink=True)
        
        data = np.load("../../Dataset/ngsim/package_multi_compact.npy")
        np.random.seed(opt.seed)
        shift = 1
        idx = np.random.permutation(min(opt.train_size + opt.test_size, len(data)))
        train_data = data[idx[:min(opt.train_size, len(data)-opt.test_size)], shift:(opt.traj_size+shift), :]
        test_data = data[idx[-opt.test_size:], shift:(opt.traj_size+shift), :]
        print(train_data.shape)
        print(test_data.shape)
        if opt.dynamic == 'Reparam':
            train_data[..., 6:8] = train_data[..., 6:8] - train_data[..., 4:6]
            test_data[..., 6:8] = test_data[..., 6:8] - test_data[..., 4:6]
        multiple_factor = [1 / np.sqrt(np.var(train_data[:, :, 6])), 1 / np.sqrt(np.var(train_data[:, :, 7]))]
        # multiple_factor = [1 / np.percentile(train_data[:, :, 6], 98), 1 / np.percentile(train_data[:, :, 7], 98)]
        train_data[..., 4] *= multiple_factor[0]
        train_data[..., 6] *= multiple_factor[0] 
        test_data[..., 4] *= multiple_factor[0]
        test_data[..., 6] *= multiple_factor[0]
        train_data[..., 5] *= multiple_factor[1]
        train_data[..., 7] *= multiple_factor[1]
        test_data[..., 5] *= multiple_factor[1]
        test_data[..., 7] *= multiple_factor[1]
        print(multiple_factor, np.min(train_data[:, :, 6]), np.max(train_data[:, :, 6]), np.min(train_data[:, :, 7]), np.max(train_data[:, :, 7]))
        # dynamic = getattr(cost_dynamic, "CarDynamic" + opt.dynamic)(opt.traj_size, multiple_factor)
        # train_data[:, :, :6] = dynamic.car_dynamic(train_data[:, 0, :6], train_data[:, :, 6:])
        # test_data[:, :, :6] = dynamic.car_dynamic(test_data[:, 0, :6], test_data[:, :, 6:])
        train_data = NGSIMSingleDatasetD(train_data, torch.FloatTensor if opt.device=="cpu" else torch.cuda.FloatTensor)
        test_data = NGSIMSingleDatasetD(test_data, torch.FloatTensor if opt.device=="cpu" else torch.cuda.FloatTensor)
        opt.multiple_factor = multiple_factor
        opt.state_size, opt.action_size, opt.extend_size = 6, 2, 47

    elif opt.data == "NGSIM_Multi":

        dataset = NGSIMMultiDataset(traj_size=opt.traj_size, max_traj=opt.train_size, agent_size_max=opt.agent_size, path="../../Dataset/ngsim/output_multi_50.pkl", 
        device=opt.device, split=4, use_original=False, 
        use_repara=(opt.dynamic=='Reparam'), shrink=True) #output_multi_201908
        # idx = np.random.permutation(min(opt.train_size + opt.test_size, len(dataset)))
        # train_data = torch.utils.data.Subset(dataset, idx[:min(opt.train_size, len(dataset)-opt.test_size)])
        # test_data = torch.utils.data.Subset(dataset, idx[-opt.test_size:])
        # train_data, test_data = torch.utils.data.random_split(dataset, [int(0.8 * len(dataset)), len(dataset) - int(0.8 * len(dataset))])
        opt.state_size, opt.action_size, opt.extend_size = 6, 2, 47
    
    else:

        raise NotImplementedError

    return train_data, test_data

def lane_function(x, c):
    res = 0
    for j in range(c.shape[-1]):
        res += c[..., j] * (x ** j)
    return res

def output_gif(x1, extend, x2=[], title='output'):
    rc('animation', html='html5')
    fig = plt.figure()
    ax = plt.gca()
    minx = 1000
    for i in range(64):
        t = np.min(x1[i, :, 0])
        if t > 0 and minx > t:
            minx = t
    maxx = np.max(x1[:, :, 0])
    ax.set_xlim((minx-10, maxx+10))
    ax.set_ylim((np.min(x1[:,:,1])-5, np.max(x1[:,:,1])+5))
    x = np.linspace(minx - 50, maxx + 50, 100)
    for c in range(64):
        if x1[c,0,0] > 0:
            ax.plot(x, lane_function(x, extend[c, :5]), c='#DDDDDD', zorder=-10, linewidth=25)

    lines = []
    for c in range(128):
        line, = ax.plot([], [], lw=1, c='r')
        lines.append(line)
        line, = ax.plot([], [], '-ok', lw=4, c='r')
        lines.append(line)
        line, = ax.plot([], [], lw=1, c='g')
        lines.append(line)
        line, = ax.plot([], [], '-ok', lw=4, c='g')
        lines.append(line)
    def init():
        for line in lines:
            line.set_data([], [])
        return (lines,)
    def animate(ii):
        def sett(l, idx, x1, x2):
            l[idx*4].set_data(x1[:ii, 0], x1[:ii, 1])
            l[idx*4+1].set_data(x1[ii-1, 0], x1[ii-1, 1])
            l[idx*4+2].set_data(x2[:ii, 0], x2[:ii, 1])
            l[idx*4+3].set_data(x2[ii-1, 0], x2[ii-1, 1])
        for c in range(64):
            if x1[c, 0, 0] > 0:
                sett(lines, c, x1[c], x2[c])
            else:
                break
        fig.savefig("%s_%d.png" % (title, ii))
        return (lines,)
    anim = animation.FuncAnimation(fig, animate, init_func=init,
                                   frames=40, interval=10, blit=False)
    anim.save("%s.gif" % (title), writer=animation.PillowWriter(fps=10))

def visualize_tray_batch_fig(x1, x2, error=None, extend=None):

    # Single-agent setting:
    fig = plt.figure(figsize=(20,10))
    fig.tight_layout()
    for i in range(min(x1.shape[0], 64)):
        ax = fig.add_subplot(8,8, i+1)
        ax.axis('off')
        ax.scatter(x1[i, :, 0], x1[i, :, 1], c='r', s=1)
        ax.scatter(x2[i, :, 0], x2[i, :, 1], c='g', s=1)
        if extend is not None and np.prod(extend.shape):
            x = np.linspace(x1[i, 0, 0], x1[i, -1, 0], 100)
            ax.plot(x, lane_function(x, extend[i, 0, -5:-1]), c='#DDDDDD', zorder=-10, linewidth=25)
            ax.plot(x, lane_function(x, extend[i, 0, -5:-1]) + 4, c='#DDDDDD', zorder=-10, linewidth=25)
            ax.plot(x, lane_function(x, extend[i, 0, -5:-1]) - 4, c='#DDDDDD', zorder=-10, linewidth=25)
        if error is not None:
            ax.title('%.4f' % (np.sum(error[0])))

    if x2.shape[2]>6:
        fig_c = plt.figure(figsize=(20, 10))
        fig_c.tight_layout()
        for i in range(min(x1.shape[0], 16)):
            ax = fig_c.add_subplot(4, 4, i+1)
            ax.axis([0,40,-2,2])
            ax.plot(x1[i, :, 6], c='#FF0000')
            ax.plot(x1[i, :, 7], c='#0000FF')
            ax.plot(x2[i, :, 6], c='#FFAAAA')
            ax.plot(x2[i, :, 7], c='#AAAAFF')
    else:
        fig_c = None
    return fig, fig_c

def visualize_tray_batch(x1, x2, title, error=None, extend=None):
    if len(x1.shape)==4:

        # Multi-agent setting 
        x1 = x1[0]
        x2 = x2[0]

        # CARLA setting
        # plt.ylim(-400, 400)
        # plt.xlim(-400, 400)
        # plt.figure(figsize=(80,80))
        plt.tight_layout()
        plt.axis('off')
        for i in range(min(50, x1.shape[0])):
            if x1[i, 0, 0]>0:
                plt.scatter(x1[i, :, 0], x1[i, :, 1], c='r', s=2)
                plt.scatter(x2[i, :, 0], x2[i, :, 1], c='g', s=2)
                # plt.text(x1[i,0,0]+2, x1[i,0,1]+2, str(i), fontsize=14)
        if extend is not None:
            for i in range(x1.shape[0]):
                if x1[i, 0, 0]>0:
                    x = np.linspace(x1[i, 0, 0] - 50, x1[i, -1, 0] + 50, 100)
                    plt.plot(x, lane_function(x, extend[i, :5]), c='#DDDDDD', zorder=-10, linewidth=25)
        if error is not None:
            plt.title('%.4f' % (np.sum(error[0])))
        plt.xlim([np.min(x1[:20,:,0]), np.max(x1[:,:,0])])
        plt.ylim([-14, 13])
        plt.savefig(title + '.png')
        # plt.savefig(title + '.eps')
        plt.clf()
        plt.close('all')

        if x2.shape[2]>6:
            plt.figure(figsize=(20, 10))
            plt.tight_layout()
            for i in range(16):
                plt.subplot(4, 4, i+1)
                plt.axis([0,40,-2,2])
                plt.plot(x1[i, :, 6], c='#FF0000')
                plt.plot(x1[i, :, 7], c='#0000FF')
                plt.plot(x2[i, :, 6], c='#FFAAAA')
                plt.plot(x2[i, :, 7], c='#AAAAFF')
            plt.savefig(title[:-4] + '_control' + title[-4:] + '.png')
            plt.clf()
            plt.close('all')

    elif len(x1.shape)==3:

        # Single-agent setting: 
        plt.figure(figsize=(20,10))
        for i in range(min(x1.shape[0], 16)):
            plt.subplot(4,4, i+1)
            plt.tight_layout()
            plt.axis('off')
            plt.scatter(x1[i, :, 0], x1[i, :, 1], c='r', s=1)
            plt.scatter(x2[i, :, 0], x2[i, :, 1], c='g', s=1)
            if extend is not None:
                x = np.linspace(x1[i, 0, 0], x1[i, -1, 0], 100)
                plt.plot(x, lane_function(x, extend[i, 0, -5:-1]), c='#DDDDDD', zorder=-10)
            if error is not None:
                plt.title('%.4f' % (np.sum(error[0])))
        plt.savefig(title + '.png')
        # plt.savefig(title + '.eps')
        plt.clf()
        plt.close('all')

        if x2.shape[2]>6:
            plt.figure(figsize=(20, 10))
            plt.tight_layout()
            for i in range(min(x1.shape[0], 16)):
                plt.subplot(4, 4, i+1)
                plt.axis([0,40,-2,2])
                plt.plot(x1[i, :, 6]*10, c='#FF0000')
                plt.plot(x1[i, :, 7], c='#0000FF')
                plt.plot(x2[i, :, 6]*10, c='#FFAAAA')
                plt.plot(x2[i, :, 7], c='#AAAAFF')
            plt.savefig(title[:-4] + '_control' + title[-4:] + '.png')
            plt.clf()
            plt.close('all')

def smooth_control(status):

    control = status[:,:,:,6:8]
    control_new = control[:,:,:-3]

def l2_error(traj_1, traj_2, squared=False):
    if squared:
        return (traj_1[..., 0] - traj_2[..., 0]) ** 2 + (traj_1[..., 1] - traj_2[..., 1]) ** 2
    return torch.sqrt((traj_1[..., 0] - traj_2[..., 0]) ** 2 + (traj_1[..., 1] - traj_2[..., 1]) ** 2)

class EvaluationMetric(object):

    def __init__(self, name):

        self.name = name
        self.reset()

    def reset(self):

        self.average_ade = []
        self.average_l2square = []
        self.minFDE = []
        self.minADE = []
        self.minRMSE = []
        self.miss_rate = []

    def calculate(self, gt, pred):
        
        gt = gt.cpu()
        pred = pred.cpu()
        if type(pred) is not list: 
            assert pred.shape[0] % gt.shape[0] == 0, "Evaluation error"
            bn, n = gt.shape[0], pred.shape[0]//gt.shape[0]
            pred = [pred[i*bn: (i+1)*bn] for i in range(n)]

        l2square = torch.stack([l2_error(gt, p, True) for p in pred]) # n * batch * traj_size
        ade = torch.stack([l2_error(gt, p, False) for p in pred])

        self.average_ade.append(ade.mean(axis=0)) # batch * traj_size
        self.average_l2square.append(l2square.mean(axis=0))
        res, pos = torch.min(ade[:, :, -1], axis=0) # batch 
        self.minFDE.append(res)
        select_best = torch.gather(ade, 0, pos.unsqueeze(1).repeat(1,gt.shape[-2]).unsqueeze(0))
        minADE = select_best.squeeze() # batch * traj_size
        self.minADE.append(minADE)
        self.miss_rate.append((res > 1).float())
        resrmse, posrmse = torch.min(l2square[:, :, -1], axis=0) # batch 
        select_best = torch.gather(l2square, 0, posrmse.unsqueeze(1).repeat(1,gt.shape[-2]).unsqueeze(0))
        minRMSE = select_best.squeeze() # batch * traj_size
        self.minRMSE.append(minRMSE)
        return self

    @staticmethod
    def _print_graph(value, tb_writer=None, name=None, epoch=None):

        if type(value) is torch.Tensor:
            value = value.cpu().data.numpy()
        fig = plt.figure()
        ax = fig.add_subplot()
        ax.plot(value)
        ax.grid(color='grey', linestyle='-', linewidth=0.5)
        for step in np.arange(10, value.shape[0], 5):
            ax.annotate("%.3f" % value[step-1], xy=(step-1, value[step-1]), ha='center', va='bottom')

        if tb_writer is None:
            return fig
        tb_writer.add_figure(name, fig, epoch)
        plt.close(fig)

    def print_save(self, tb_writer, epoch):

        rmse = torch.sqrt(torch.mean(torch.cat(self.average_l2square), dim=0))
        tb_writer.add_figure("%s/avgRMSE_time" % self.name, self._print_graph(rmse), epoch)
        minADE = torch.mean(torch.cat(self.minADE), dim=0)
        # tb_writer.add_figure("%s/minADE_time" % self.name, self._print_graph(minADE), epoch)
        avgADE = torch.mean(torch.cat(self.average_ade), dim=0)
        # tb_writer.add_figure("%s/avgADE_time" % self.name, self._print_graph(avgADE), epoch)
        minRMSE = torch.sqrt(torch.mean(torch.cat(self.minRMSE), dim=0))
        # tb_writer.add_figure("%s/minRMSE_time" % self.name, self._print_graph(minRMSE), epoch)

        res_rmse, res_ade, res_fde = minRMSE[-1], torch.mean(minADE), torch.mean(torch.cat(self.minFDE), dim=0)
        tb_writer.add_scalar("%s/avgFDE" % self.name, avgADE[-1], epoch)
        tb_writer.add_scalar("%s/L2square" % self.name, torch.mean(torch.cat(self.average_l2square), dim=0)[-1], epoch)
        tb_writer.add_scalar("%s/rmse_e" % self.name, rmse[-1], epoch)
        if len(rmse) > 30:
            tb_writer.add_scalar("%s/rmse_3" % self.name, rmse[29], epoch)
        tb_writer.add_scalar("%s/minFDE" % self.name, res_fde, epoch)
        tb_writer.add_scalar("%s/minRMSE" % self.name, res_rmse, epoch)
        tb_writer.add_scalar("%s/minADE" % self.name, res_ade, epoch)
        tb_writer.add_scalar("%s/miss_rate" % self.name, torch.mean(torch.cat(self.miss_rate), dim=0), epoch)
        self.reset()
        return float(res_rmse), float(res_ade), float(res_fde)

def traj_likelihood(traj_1, traj_2):

    if len(traj_1.shape) == 2:
        scale_0 = 5
        scale_1 = 1.5
        likelihood = 1
        prob_0 = norm.pdf(0, 0, scale_0)
        prob_1 = norm.pdf(0, 0, scale_1)
        for pt_1, pt_2 in zip(traj_1, traj_2):
            p_0 = norm.pdf(pt_1[0], loc=pt_2[0], scale=scale_0)
            p_1 = norm.pdf(pt_1[1], loc=pt_2[1], scale=scale_1)
            likelihood *= p_0 * p_1 / (prob_0 * prob_1)
        return likelihood
    else:
        n = traj_1.shape[0]
        scale_0 = 5
        scale_1 = 1.5
        likelihood = np.ones(n)
        prob_0 = norm.pdf(0, 0, scale_0)
        prob_1 = norm.pdf(0, 0, scale_1)
        for i in range(n):
            for t in range(traj_1.shape[1]):
                p_0 = norm.pdf(traj_1[i, t, 0], loc=traj_2[i, t, 0], scale=scale_0)
                p_1 = norm.pdf(traj_1[i, t, 1], loc=traj_2[i, t, 1], scale=scale_1)
                likelihood[i] *= p_0 * p_1 / (prob_0 * prob_1)
        return likelihood

def basic_check(status_long, car_dynamic, config):

    status = status_long[..., config.his_size-1:config.traj_size, :]
    print(status.shape)
    # Test gt u:
    u = status[..., -2:]
    total_vehicle = torch.sum(status[..., 0, 0] != 0)
    print("%d Total vehicle in %d scene" % (total_vehicle, status.shape[0]))
    x_new = car_dynamic(status, u)
    l2_baseline = torch.sqrt(torch.sum(l2_error(status, x_new)[..., -1] ** 2) / total_vehicle).cpu().data.numpy()
    print("Reference result for gt control: %.3f" % l2_baseline)
    # Test gt u from x:
    if len(status.shape)==4:
        u_gt = torch.cat(((status[..., 1:, 2:4] - status[..., :-1, 2:4])*10,
                   torch.zeros((status.shape[0], status.shape[1], 1, 2), device=status.device)), 2)
    else:
        u_gt = torch.cat(((status[..., 1:, 2:4] - status[..., :-1, 2:4]) * 10,
                          torch.zeros((status.shape[0], 1, 2), device=status.device)), 1)

    x_new = car_dynamic(status, u_gt)
    l2_baseline = torch.sqrt(torch.sum(l2_error(status, x_new)[..., -1] ** 2) / total_vehicle).cpu().data.numpy()
    print("Reference result for gt from u control: %.3f" % l2_baseline)
    u = torch.zeros_like(status[..., -2:], device=status.device)
    x_new = car_dynamic(status, u)

def basic_check_np(status, car_dynamic):

    # Test gt u:
    status = status.cpu().data.numpy()
    u = status[..., 6:8]
    total_vehicle = np.sum(status[:, :, 10, 0] != 0)
    print("%d Total vehicle in %d scene", (status.shape[0], total_vehicle))
    x_new = car_dynamic(status, u)
    l2_baseline = np.sqrt(np.sum(l2_error(status, x_new)[..., -1] ** 2) / total_vehicle)
    print("Reference result for gt control: %.3f" % l2_baseline)


class LogSaver(object):

    def __init__(self, num_epochs, traj_size, self_only=False, save_theta=None, tb_writer=None):

        self.train_log = {
            'number' : np.zeros(num_epochs+1),
            'likelihood' : np.zeros(num_epochs+1),
            'l2norm' : np.zeros((num_epochs+1, traj_size)),
            'rmse' : np.zeros((num_epochs+1, traj_size)),}
        self.test_log = {
            'number' : np.zeros(num_epochs+1),
            'likelihood' : np.zeros(num_epochs+1),
            'l2norm' : np.zeros((num_epochs+1, traj_size)),
            'rmse' : np.zeros((num_epochs+1, traj_size)),  
            'gen_loss' : np.zeros(num_epochs+1), 
            'des_loss' : np.zeros(num_epochs+1)}
        self.general = {
            'save_theta' : (save_theta is not None),
            'theta' : None
        }
        if save_theta is not None:
            self.general['theta'] = np.zeros((num_epochs + 1, 10))
        self.single_only = self_only
        self.tb_writer = SummaryWriter(log_dir=tb_writer) if type(tb_writer) is str else tb_writer 

    def save(self, i, likelihood, l2norm, train=True, theta=None):

        pnt = self.train_log if train else self.test_log
        if self.single_only:
            pnt['number'][i] += l2norm.shape[0]
            pnt['likelihood'][i] += np.sum(likelihood**2)
            pnt['l2norm'][i] += np.sum(l2norm[:, :], axis=0)
            pnt['rmse'][i] += np.sum((l2norm**2)[:, :], axis=0)
        else:
            pnt['number'][i] += np.sum(l2norm[:, :, -1] > 0.001)
            pnt['likelihood'][i] += np.sum(likelihood**2)
            pnt['l2norm'][i] += np.sum(l2norm, axis=(0, 1))
            pnt['rmse'][i] += np.sum(l2norm**2, axis=(0, 1))
        if theta is not None:
            self.general['theta'][i] = theta

    @property
    def test_rmse(self):
        return [np.sqrt(self.test_log['rmse'][epoch, 29 if self.test_log['rmse'].shape[1]>30 else -1] /
        self.test_log['number'][epoch]) for epoch in range(len(self.test_log['number'])-1)]

    def mesg(self, epoch, batch=-1):

        mesg = 'Epoch #%d' % epoch
        if batch>=0:
            mesg += '-%d' % batch

        a = self.train_log['number'][epoch] if self.train_log['number'][epoch]>0 else 1e10
        b = self.test_log['number'][epoch] if self.test_log['number'][epoch]>0 else 1e10
        # print(np.sqrt(self.test_log['rmse'][epoch, :]/b))
        mesg += ': gen_RMSE = %.3f / %.3f ; ADE : %.3f / %.3f ; RMSE : %.3f / %.3f' \
                % (np.sqrt(self.train_log['likelihood'][epoch]/a), np.sqrt(self.test_log['likelihood'][epoch]/b),
                   self.train_log['l2norm'][epoch, -1]/a, self.test_log['l2norm'][epoch, -1]/b,
                   np.sqrt(self.train_log['rmse'][epoch, -1]/a), np.sqrt(self.test_log['rmse'][epoch, -1]/b))
        return mesg

    def output_image(self, epoch, output_dir):
        
        idxs = [e for e,i in enumerate(self.train_log['number']) if i!=0]
        a = self.train_log['number'][epoch]
        b = self.test_log['number'][epoch]
        plt.plot(idxs, self.train_log['likelihood'][idxs]/a, label='Training Likelihood')
        plt.plot(idxs, self.test_log['likelihood'][idxs]/b, label='Testing Likelihood')
        plt.legend()
        plt.savefig(output_dir + '/likelihood.png')
        plt.clf()
        plt.plot(idxs, np.sqrt(self.train_log['rmse'][idxs, -1]/a), label='Training RMSE')
        plt.plot(idxs, np.sqrt(self.test_log['rmse'][idxs, -1]/b), label='Testing RMSE')
        plt.legend()
        plt.savefig(output_dir + '/l2norm.png')
        plt.clf()
        if self.general['save_theta']:
            plt.plot(idxs, self.general['theta'][idxs], label='a')
            plt.legend()
            plt.savefig(output_dir + '/learned_theta.png')
            plt.clf()

    def tb_step(self, e):
        a = self.train_log['number'][e]
        b = self.test_log['number'][e]
        self.tb_writer.add_scalar("train/gen_rmse", self.train_log['likelihood'][e] / a, e)
        self.tb_writer.add_scalar("train/rmse", self.train_log['rmse'][e, -1] / a, e)
        self.tb_writer.add_scalar("train/l2norm", self.train_log['l2norm'][e, -1] / a, e)
        self.tb_writer.add_scalar("test/gen_rmse", self.test_log['likelihood'][e] / a, e)
        self.tb_writer.add_scalar("test/rmse", self.test_log['rmse'][e, -1] / a, e)
        self.tb_writer.add_scalar("test/l2norm", self.test_log['l2norm'][e, -1] / a, e)
        if self.general['save_theta']:
            for i, v in enumerate(self.general['theta'][e]):
                self.tb_writer.add_scalar("general/theta_%d" % i, v, e)
                
    def save_dict(self, theta, path):

        save_dict = {'train_l2' : self.train_log['l2norm'],
                     'test_l2' : self.test_log['l2norm'],
                     'train_rmse' : self.train_log['rmse'],
                     'test_rmse' : self.test_log['rmse'],
                     'train_li' : self.train_log['likelihood'],
                     'test_li' : self.test_log['likelihood'],
                     'train_number' : self.train_log['number'],
                     'test_number' : self.test_log['number'],
                     'theta' : theta
                     }
        np.save(path, save_dict)



def save_config(config, add_message=None):

    # Save to all
    with open('output/experiment_list.csv', "a") as output:
        writer = csv.writer(output)
        dd = time.strftime("%Y-%m-%d %H:%M:%S", time.localtime())
        row = [dd, str(config)]
        if add_message is not None:
            row = row + [add_message]
        writer.writerow(row)

    # Save full config

    log_path = os.path.join('output', config.category)
    if not os.path.exists(log_path):
        os.makedirs(log_path)

    dict_config = vars(config)
    with open(log_path + '/config.cfg', "w") as output:
        writer = csv.writer(output)
        for val in dict_config:
            writer.writerow([val + " : " + str(dict_config[val])])
    with open(log_path + "/config.pkl", "wb") as conf:
        pickle.dump(config, conf)
